"""
Download / Cache Utilities
"""

from pathlib import Path
from hashlib import md5
from requests import get
from tqdm import tqdm

CACHE_HOME = Path.home() / ".camtrapml"


def download(url: str, path: Path, md5_hash: str) -> None:
    """
    Downloads a file from a URL to a path.
    """

    if path.exists() and (md5_hash == "" or md5_hash == hash_file(path)):
        return

    path.parent.mkdir(parents=True, exist_ok=True)

    resp = get(url, stream=True)
    total = int(resp.headers.get("content-length", 0))

    with path.open("wb") as file, tqdm(
        desc="Downloading Model",
        total=total,
        unit="iB",
        unit_scale=True,
        unit_divisor=1024,
    ) as progress_bar:
        for data in resp.iter_content(chunk_size=1024):
            size = file.write(data)
            progress_bar.update(size)

    if md5_hash != "":
        assert md5_hash == hash_file(path)


def hash_file(path: Path) -> str:
    """
    Hashes a file.
    """

    with open(path, "rb") as file_handle:
        return md5(file_handle.read()).hexdigest()



"""
Detection Utilities
"""
from typing import Union
from PIL import Image, ImageDraw, ImageFont
from font_fredoka_one import FredokaOne  # pylint: disable=E0611
from ..image.utils import ImageSource, load_image

def render_detections(
    image_source: ImageSource,
    detections,
    draw_box=True,
    draw_score=True,
    class_map: Union[dict, None] = None,
):
    """
    Renders the image with the detections.

    Args:

      image_path: Path to the image to render the detections on.
      detections: List of tuples containing the bounding box coordinates and the
        class of the detected object.

    Returns:

      The rendered image as PIL.Image.
    """

    image = load_image(image_source)
    draw = ImageDraw.Draw(image)

    for detection in detections:
        min_y, min_x, max_y, max_x = detection["bbox"]

        if draw_box:
            draw.rectangle(
                (
                    (min_x * image.size[0]),
                    (min_y * image.size[1]),
                    (max_x * image.size[0]),
                    (max_y * image.size[1]),
                ),
                outline="red",
                width=3,
            )

        if class_map or draw_score:
            font = ImageFont.FreeTypeFont(FredokaOne, size=40)
            class_name = (
                class_map[detection["category"]]
                if class_map and detection["category"] in class_map
                else str(detection["category"])
            )
            text = class_name if class_map else ""
            text += f" {detection['conf']:.2f}" if draw_score else ""
            draw.text(
                (
                    (min_x * image.size[0]),
                    (min_y * image.size[1]) - 50,
                ),
                text,
                fill="red",
                font=font,
            )

    return image


def remove_detections_from_image(image: Image.Image, detections):
    """
    Removes the detections from an image.

    Args:

      image: Image to remove the detections from.
      detections: List of detections to remove.

    Returns:

      The rendered image as PIL.Image.
    """

    image = image.copy()
    image = image.convert("RGB")

    for detection in detections:
        min_y, min_x, max_y, max_x = detection["bbox"]
        image_width, image_height = image.size

        cover_size = (
            int(image_width * (max_x - min_x)),
            int(image_height * (max_y - min_y)),
        )
        cover_position = (int(image_width * min_x), int(image_height * min_y))

        cover = Image.new("RGBA", cover_size, (0, 0, 0, 0))

        image.paste(cover, cover_position)

    return image


def extract_detections_from_image(image: Image.Image, detections):
    """
    Extracts the detections from an image.

    Args:

      image: Image to extract the detections from.
      detections: List of detections to extract.

    Returns:

      The rendered image as PIL.Image.
    """

    image_width, image_height = image.size

    for detection in detections:
        min_y, min_x, max_y, max_x = detection["bbox"]
        image_width, image_height = image.size

        yield image.crop(
            (
                (min_x * image_width),
                (min_y * image_height),
                (max_x * image_width),
                (max_y * image_height),
            )
        )



"""
General image processing functions.
"""

from typing import Tuple, Union
from pathlib import Path
from urllib.parse import urlparse
from io import BytesIO
from PIL import Image
from requests import get

ImageSource = Union[Image.Image, Path, str]


def load_image(source: ImageSource, mode="RGB", timeout=10) -> Image.Image:
    """
    Loads an image from a path or url or returns a copy of a previously loaded image.
    """

    if isinstance(source, str) and urlparse(source).scheme in ["http", "https"]:
        image = Image.open(BytesIO(get(source, timeout=timeout).content))

    elif isinstance(source, Image.Image):
        image = source.copy()

    else:
        image = Image.open(source)

    if mode is not None:
        image = image.convert(mode)

    return image


def thumbnail(source: ImageSource, size: Tuple[int, int] = (400, 400)) -> Image.Image:
    """
    Generates a thumbnail of an image keeping aspect ratio. The size argument
    is the maximum size of the thumbnail.
    """
    image = load_image(source)
    image.thumbnail(size, Image.ANTIALIAS)
    return image


def is_image(path: ImageSource) -> bool:
    """
    Checks if a path is an image, or at least one that can be read with Pillow.
    """
    try:
        load_image(path)
        return True
    except Exception:  # pylint: disable=W0703
        return False



"""
MegaDetector, developed by Microsoft and collaborators aims to be a
generlisable detection model for detecting animals, humans and vehicles in
camera trap imagery.

More information can be found on Microsoft's
[CameraTrap GitHub Repository](https://github.com/microsoft/CameraTraps/blob/main/megadetector.md).

This module contains just enough code to run each of MegaDetector versions
where each version is provided with a consistent API.


    from camtrapml.detection.models.megadetector import MegaDetectorV4_1

    with MegaDetectorV4_1() as detector:
        # Run detection on a single image
        detections = detector.detect(image)

"""
from pathlib import Path
from typing import Tuple, Union
from json import load
from .tensorflow import TF1ODAPIFrozenModel
from ...download import CACHE_HOME

MODEL_FILES_ROUTE = "https://lilablobssc.blob.core.windows.net/models/camera_traps/megadetector/"


class MegaDetectorV4_1(TF1ODAPIFrozenModel):  # pylint: disable=C0103
    """
    MegaDetector v4.1
    """

    class_map = {
        1: "animal",
        2: "human",
        3: "vehicle",
    }

    model_name = "megadetector"
    model_version = "v4.1.0"
    model_url = MODEL_FILES_ROUTE + "md_v4.1.0/md_v4.1.0.pb"
    model_hash = "7008707c66787a4d1c6544c27523f714"
    model_path = CACHE_HOME / "models" / "megadetector" / "v4.1.0" / "md_v4.1.0.pb"


class MegaDetectorV3(TF1ODAPIFrozenModel):
    """
    MegaDetector v3.0
    """

    class_map = {
        1: "animal",
        2: "human",
    }

    model_name = "megadetector"
    model_version = "v3.0.0"
    model_url = MODEL_FILES_ROUTE + "megadetector/megadetector_v3.pb"
    model_hash = "1a4ca31abf9dca286ff1641eedbccc3d"
    model_path = CACHE_HOME / "models" / "megadetector" / "v3.0.0" / "md_v3.0.0.pb"


class MegaDetectorV2(TF1ODAPIFrozenModel):
    """
    MegaDetector v2.0
    """

    class_map = {
        1: "animal",
    }

    model_name = "megadetector"
    model_version = "v2.0.0"
    model_url = MODEL_FILES_ROUTE + "/megadetector/megadetector_v2.pb"
    model_hash = "3a634a610d9a74f671cef08f75cfd661"
    model_path = CACHE_HOME / "models" / "megadetector" / "v2.0.0" / "md_v2.0.0.pb"


def read_megadetector_batch_file(
    path: Union[Path, str], image_dir
) -> Tuple[list, list]:
    """
    Reads a batch file from a MegaDetector model.

    Args:
        path: Path to the batch file.
        image_dir: Path to the directory containing the images.

    Returns:
        A tuple of (image_paths, image_detections)
    """

    image_paths = []
    image_detections = []

    with open(path, "r", encoding="UTF-8") as file_handle:
        images = load(file_handle)["images"]

    for image_data in images:
        image_path = image_dir / image_data["file"]
        image_paths.append(image_path)
        image_detections.append(image_data["detections"])

    return image_paths, image_detections



"""
Dataset Handling
"""

from typing import Union
from pathlib import Path
from os import walk
from PIL import Image
from camtrapml.image.utils import is_image


class ImageDataset:
    """
    A dataset is a collection of images.
    """

    name: str
    path: Path

    def __init__(self, path: Path, name: Union[None, str] = None):
        self.name = name
        self.path = Path(path).expanduser()
        self.image_set = None
        self.detection_set = None

        if not self.path.exists() or not self.path.is_dir():
            raise ValueError(f"{self.path} is not a directory")

    def enumerate_images(self, enhanced_validation: bool = False):
        """
        Enumerates all images in the dataset.
        """

        supported_extensions = {
            extention.lower() for extention, image_format in
            Image.registered_extensions().items()
            if image_format in Image.OPEN
        }

        for root, _, files in walk(self.path):
            for file in files:
                file_path = Path(root) / file

                if enhanced_validation:
                    if is_image(file_path):
                        yield file_path

                else:
                    if file_path.suffix.lower() in supported_extensions:
                        yield file_path

    @staticmethod
    def from_coco(source):
        """
        Load a Dataset Instance from COCO-CameraTrap
        """
        raise NotImplementedError()



"""
Tensorflow Detection Model Base Classes
"""

from platform import system, machine
from warnings import warn
from typing import Union
from pathlib import Path

import numpy as np
import tensorflow as tf

from camtrapml.image.utils import ImageSource, load_image
from camtrapml.download import download, hash_file


class TF1ODAPIFrozenModel:
    """
    Tensorflow v1 (loaded using v2) Object Detection API Frozen Model Loader
    """

    class_map = None

    model_path = None
    model_version = None
    model_url = ""

    def __init__(
        self,
        model_path: Union[Path, str, None] = None,
        model_url: str = "",
        model_hash: str = "",
        class_map: Union[dict, None] = None,
    ) -> None:
        """
        Initialises a Tensorflow V1 Object Detection Frozen Graph Model.

        Args:

          model_path: Path to the frozen model .pb file.
        """

        if model_path:
            self.model_path = model_path

        if model_url:
            self.model_url = model_url

        if model_hash:
            self.model_hash = model_hash

        if class_map:
            self.class_map = class_map

        self._session = None
        self._tensors = None

        if Path(self.model_path).exists() is False and self.model_url != "":
            download(self.model_url, self.model_path, "")
        elif (
            Path(self.model_path).exists() is True
            and model_hash in dir(self)
            and self.model_hash != ""
        ):
            local_hash = hash_file(self.model_path)
            if local_hash != self.model_hash:
                raise ValueError(
                    f"Hash mismatch for model {self.model_path}. Local hash: {local_hash}."
                )

    def load_model(self,) -> None:
        """
        Loads the model from the frozen model .pb file.
        """

        # Disable the GPU on M1 Macs as there appears to be a bug where the outputs are
        # not the same as the reference implementation.
        if system() == "Darwin" and machine() == "arm64":
            warn("Disabling GPU on M1 Macs")
            tf.config.set_visible_devices([], "GPU")

        detection_graph = tf.compat.v1.Graph()

        with detection_graph.as_default():  # pylint: disable=E1129
            od_graph_def = tf.compat.v1.GraphDef()
            with tf.compat.v1.gfile.GFile(str(self.model_path), "rb") as fid:
                serialized_graph = fid.read()
                od_graph_def.ParseFromString(serialized_graph)
                tf.import_graph_def(od_graph_def, name="")

        self._session = tf.compat.v1.Session(graph=detection_graph)

        # Get the tensors we need.
        self._tensors = {
            'image': detection_graph.get_tensor_by_name(
                "image_tensor:0"
            ),
            'boxes': detection_graph.get_tensor_by_name(
                "detection_boxes:0"
            ),
            'scores': detection_graph.get_tensor_by_name(
                "detection_scores:0"
            ),
            'classes': detection_graph.get_tensor_by_name(
                "detection_classes:0"
            )
        }

    def close(self) -> None:
        """
        Closes the model.
        """
        if self._session:
            self._session.close()

        self._session = None
        self._tensors = None

    def detect(self, image_source: ImageSource, min_score: float = 0.1) -> list:
        """
        Detects objects in the image.

        Args:

          image_source: PIL Image, Path to image or URL to image.
          min_score: Minimum score to consider a detection.

        Returns:

          A list of tuples containing the bounding box coordinates and the class
          of the detected object.
        """

        # Load the detection graph once.
        if self._session is None:
            self.load_model()

        # Load the image.
        image_np = load_image(image_source, mode="RGB")

        # Expand dimensions since the model expects images to have shape:
        # [batch_size, None, None, 3]
        image_np_expanded = np.expand_dims(image_np, axis=0)

        (box_tensor_out, score_tensor_out, class_tensor_out) = self._session.run(
            [
                self._tensors['boxes'],
                self._tensors['scores'],
                self._tensors['classes'],
            ],
            feed_dict={self._tensors['image']: image_np_expanded},
        )

        # Filter out the low scoring detections and return the bounding boxes.
        boxes = []
        for i in range(len(box_tensor_out[0])):
            if score_tensor_out[0][i] > min_score:
                boxes.append(
                    {
                        "category": int(class_tensor_out[0][i]),
                        "conf": score_tensor_out[0][i],
                        "bbox": box_tensor_out[0][i].tolist(),
                    }
                )

        return boxes

    def __enter__(self):
        self.load_model()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



"""
Tools for extracting EXIF data from images.
"""

from pathlib import Path
from typing import List, Union
from exiftool import ExifTool
from tqdm.contrib.concurrent import process_map


def extract_exif(path: Union[Path, str]) -> dict:
    """
    Extracts EXIF data from an image.
    """
    with ExifTool() as exiftool:
        return exiftool.get_metadata(filename=str(path))


def extract_multiple_exif(paths: List[Union[Path, str]]) -> list:
    """
    Extracts EXIF data from multiple images.
    """
    with ExifTool() as exiftool:
        return exiftool.get_metadata_batch([str(path) for path in paths])


def extract_multiple_exif_fast(paths: List[Union[Path, str]], batch_size=100) -> list:
    """
    Extracts EXIF data from a list of images by spreading the load across
    multiple processes making use of multiple CPU cores.
    """
    if len(paths) <= batch_size:
        batch_size = len(paths)

    paths = [str(path) for path in paths]
    batched_paths = list(zip(*[iter(paths)] * batch_size))

    exif_data = process_map(
        extract_multiple_exif, batched_paths, desc="Extracting EXIF data", chunksize=3
    )

    # Unbatch
    return [item for sublist in exif_data for item in sublist]

def parse_iptc(keywords) -> dict:
    """
    Parses IPTC Tags
    """

    if isinstance(keywords, str):
        keywords = [keywords]

    def parse(keyword):
        split = keyword.split(':')
        if len(split) == 1:
            return split[0], True

        key = split[0].strip()
        val = split[1].strip()

        return key, val

    return [parse(keyword) for keyword in keywords]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['camtrapml',
 'camtrapml.detection',
 'camtrapml.detection.models',
 'camtrapml.image']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9,<10',
 'PyExifTool>=0.4.13,<0.5.0',
 'font-fredoka-one>=0.0.4,<0.0.5',
 'h5py>=3.7.0,<4.0.0',
 'matplotlib>=3,<4',
 'numpy>=1,<2',
 'requests>=2,<3',
 'scikit-learn==1.0.1',
 'scipy>=1.9.1,<2.0.0',
 'tensorflow-hub>=0.12.0,<0.13.0',
 'tensorflow>=2.10.0,<3.0.0',
 'tqdm>=4,<5']

setup_kwargs = {
    'name': 'camtrapml',
    'version': '0.4.1',
    'description': 'CamTrapML is a Python library for Detecting, Classifying, and Analysing Wildlife Camera Trap Imagery.',
    'long_description': '# CamTrapML\n\n> CamTrapML is a Python library for Detecting, Classifying, and Analysing Wildlife [Camera Trap](https://en.wikipedia.org/wiki/Camera_trap) Imagery.\n\n## Installation\n\n    $ pip install camtrapml\n\n## Features\n\n### Loading Data\n\nSearch for images in a directory, load an image and create a thumbnail.\n\n\n```python\n%load_ext autoreload\n%autoreload\n\nfrom camtrapml.dataset import ImageDataset\nfrom camtrapml.image.utils import load_image, thumbnail\n\nimageset = ImageDataset(\n    name="Test Images",\n    path="test/fixtures/images",\n)\n\nimage_paths = list(imageset.enumerate_images())\n\nthumbnail(load_image(image_paths[0]))\n```\n\n### EXIF Extraction\n\nEXIF extraction is a common task in gathering the metadata such as each image\'s timestamp, camera model, focal length, etc. Some researchers write labelling into the EXIF data. CamTrapML doesn\'t assist with writing to EXIF. However, there is functionality for extracting it for analysis and building datasets for training new models from previously labelled images.\n\nExifTool is required for this package to work. Installation instructions can be found [here](https://exiftool.org/install.html).\n\nThree methods are available for extracting EXIF data from images. Each with different performance characteristics.\n\n**Method 1: Individual Images**\n\n\n```python\nfrom camtrapml.image.exif import extract_exif\n\nexif = extract_exif(image_paths[0])\nexif\n```\n\n**Method 2: Multiple Images**\n\n`extract_multiple_exif` passes a list of image paths to ExifTool and returns a list of dictionaries containing the EXIF data. This is faster than `extract_exif` when multiple images are being processed as it only passes the list of image paths to ExifTool once, rather than spawning a new process for each image.\n\n\n```python\nfrom camtrapml.image.exif import extract_multiple_exif\n\nexif = extract_multiple_exif(image_paths)\nexif[0]\n```\n\n**Method 3: Multiple Images, Multiple Processes**\n\nWhen processing large datasets, it\'s apparent that the bottleneck in extracting the EXIF information tends to be the CPU. This method spawns multiple versions of ExifTool in parallel, each with a batch of image paths. This is faster than `extract_multiple_exif` when processing large datasets as it allows for multiple processes to be spawned and the data extracted in parallel.\n\n\n```python\nfrom camtrapml.image.exif import extract_multiple_exif_fast\n\nexif = extract_multiple_exif_fast(image_paths)\nexif[0]\n```\n\n### Detection\n\nVarious Detection models are available in the `camtrapml.detection` subpackage. These currently include MegaDetector (v4.1, v3 and v2) and support for loading in custom Tensorflow v1.x Object Detection Frozen models.\n\n#### Detection with MegaDetector v4.1\n\n\n```python\nfrom camtrapml.detection.models.megadetector import MegaDetectorV4_1\nfrom camtrapml.detection.utils import render_detections\n\nwith MegaDetectorV4_1() as detector:\n    detections = detector.detect(image_paths[0])\n\nthumbnail(\n    render_detections(image_paths[0], detections, class_map=detector.class_map)\n)\n```\n\n#### Detection with a custom Tensorflow v1.x Object Detection Frozen model\n\n\n```python\n!cp ~/.camtrapml/models/megadetector/v4.1.0/md_v4.1.0.pb example-custom-model.pb\n\nfrom camtrapml.detection.models.tensorflow import TF1ODAPIFrozenModel\nfrom camtrapml.detection.utils import render_detections\nfrom pathlib import Path\n\nwith TF1ODAPIFrozenModel(\n    model_path=Path("example-custom-model.pb").expanduser(),\n    class_map={\n        1: "animal",\n    },\n) as detector:\n    detections = detector.detect(image_paths[1])\n\nthumbnail(\n    render_detections(image_paths[1], detections, class_map=detector.class_map)\n)\n```\n\n#### Extract Detections\n\n\n```python\nfrom camtrapml.detection.models.megadetector import MegaDetectorV4_1\nfrom camtrapml.detection.utils import extract_detections_from_image\n\nwith MegaDetectorV4_1() as detector:\n    detections = detector.detect(image_paths[0])\n\nlist(extract_detections_from_image(load_image(image_paths[0]), detections))[0]\n```\n\n#### Remove Humans from Images\n\nIn order to reduce the risks of identification of humans in line with GDPR, CamTrapML provides the ability to remove humans from images. This is achieved by using the MegaDetector v3+ models to detect humans in the image, and then replacing all pixels in each human detection.\n\n\n```python\nfrom camtrapml.detection.models.megadetector import MegaDetectorV4_1\nfrom camtrapml.detection.utils import remove_detections_from_image\nfrom camtrapml.image.utils import load_image, thumbnail\nfrom pathlib import Path\n\nct_image_with_humans = Path("test/fixtures/human_images/IMG_0254.JPG").expanduser()\n\nwith MegaDetectorV4_1() as detector:\n    detections = detector.detect(ct_image_with_humans)\n\nhuman_class_id = 2\n\nthumbnail(\n    remove_detections_from_image(\n        load_image(ct_image_with_humans),\n        [\n            detection\n            for detection in detections\n            if detection["category"] == human_class_id and detection["conf"] > 0.5\n        ],\n    )\n)\n```\n\n\n```python\n\n```\n',
    'author': 'Benjamin C. Evans',
    'author_email': 'Benjamin.Evans@brunel.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/bencevans/camtrapml',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
